#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ZipArchive/ZipArchive.h"

#define TEXMOD_XOR 0x3FA43FA4

const char texmod_password[] = {0x73, 0x2A, 0x63, 0x7D, 0x5F, 0x0A, 0xA6, 0xBD,
								0x7D, 0x65, 0x7E, 0x67, 0x61, 0x2A, 0x7F, 0x7F,
								0x74, 0x61, 0x67, 0x5B, 0x60, 0x70, 0x45, 0x74,
								0x5C, 0x22, 0x74, 0x5D, 0x6E, 0x6A, 0x73, 0x41,
								0x77, 0x6E, 0x46, 0x47, 0x77, 0x49, 0x0C, 0x4B,
								0x46, 0x6F, '\0'
							   };

// TODO: check for exceptions like failure to open/create file
int main( int args, char* argv[] )
{
	char *input, *output;

	if ( args != 2 && args != 3 )
	{
		printf(
			"Converts a .tpf file created by TexMod into a standard ZIP file.\n"
			"\n"
			"opentpf.exe input [output]\n"
			"\n"
			"  input        TPF file created by TexMod.\n"
			"  output       Output file. Defaults to input with '.zip' appended.\n"
			"\n" );
		return 0;
	}

	input = argv[1];

	if ( args == 3 )
	{
		output = argv[2];
	}
	else
	{
		output = ( char* )malloc( strlen( input ) + 4 );
		strcpy( output, input );
		strcat( output, ".zip" );
	}

	try {
		CZipMemFile memFile;


		// Copy file to *data
		CZipFile fin;
		if( !fin.Open( input , CZipFile::modeRead, false ) )
		{
			printf("Failed to open input file: %s", input );
			return 0;
		}
		ZIP_FILE_USIZE length = fin.GetLength();
		BYTE* data = ( BYTE* )malloc( length );
		if ( !data )
		{
			printf("Failed to allocate memory\n");
			return 0;
		}
		fin.Read( data, length );
		fin.Close();


		// XOR data
		ZIP_FILE_USIZE pos = 0;
		unsigned int operand = TEXMOD_XOR;

		// 4 bytes at a time first
		while ( pos <= length - 4 )
		{
			*( unsigned int* )( &data[pos] ) ^= operand;
			pos += 4;
		}
		// last extra bytes
		while ( pos < length )
		{
			data[pos] ^= ( char )( operand >> 24 );
			operand <<= 4;
			pos++;
		}
		memFile.Attach( data, length );


		// Decrypt files, saving them to outZip
		CZipArchive inZip;
		inZip.Open( memFile, CZipArchive::zipOpenReadOnly );
		inZip.SetPassword( texmod_password );

		CZipArchive outZip;
		if( !outZip.Open( output, CZipArchive::zipCreate ) )
		{
			printf("Failed to create output file: %s", output );
			return 0;
		}
		for ( ZIP_INDEX_TYPE i = 0; i < inZip.GetCount(); i++ )
		{
			inZip.OpenFile( i );
			DWORD uSize = ( DWORD )inZip[i]->m_uUncomprSize;
			LPCTSTR name = inZip[i]->GetFileName();
			printf( "%d of %d\t%s\n", i + 1, inZip.GetCount(), name );
			BYTE *uData = ( BYTE* ) malloc( uSize );
			if ( !uData )
			{
				printf("Failed to allocate memory\n");
				return 0;
			}
			inZip.ReadFile( uData, uSize );
			inZip.CloseFile();

			CZipMemFile memFile( uData, uSize );
			outZip.AddNewFile( memFile, name );

			free( uData );
		}
		inZip.Close();
		outZip.Close();


		// dealloc memory, not really necessary //

		// memFile was previously detached so we have to free it
		free( memFile.Detach() );

	}
	catch( CZipException e )
	{
		printf("ZipArchive error: %s",(LPCTSTR)(e.GetErrorDescription()));
	}

	if ( args != 3 )
		free( output );

	return 0;
}
